from typing import (
    Any,
    Dict,
    List,
    Optional,
)

from easypost.requestor import Requestor
from easypost.resource import (
    AllResource,
    CreateResource,
)


class Tracker(AllResource, CreateResource):
    @classmethod
    def create_list(cls, trackers: List[Dict[str, Any]], api_key: Optional[str] = None) -> bool:
        """Create a list of trackers."""
        requestor = Requestor(local_api_key=api_key)
        url = "%s/%s" % (cls.class_url(), "create_list")
        new_params = {"trackers": trackers}
        _, _ = requestor.request(method="post", url=url, params=new_params)
        return True



VERSION = "7.0.0"

elements = VERSION.split("-")
numbers = [str(v) for v in elements[0].split(".")]
if len(elements) > 1:
    numbers.extend(elements[1])
numbers.extend(elements[1:])
VERSION_INFO = numbers



import json

import easypost
from easypost.easypost_object import convert_to_easypost_object
from easypost.resource import (
    AllResource,
    Resource,
)


class Event(AllResource, Resource):
    @classmethod
    def receive(cls, values: str) -> "Event":
        return convert_to_easypost_object(response=json.loads(s=values), api_key=easypost.api_key)



from typing import Optional

from easypost.easypost_object import convert_to_easypost_object
from easypost.requestor import Requestor
from easypost.resource import (
    AllResource,
    CreateResource,
)


class Batch(AllResource, CreateResource):
    @classmethod
    def create_and_buy(cls, api_key: Optional[str] = None, **params) -> "Batch":
        """Create and buy a Batch."""
        requestor = Requestor(local_api_key=api_key)
        url = "%s/%s" % (cls.class_url(), "create_and_buy")
        wrapped_params = {cls.snakecase_name(): params}
        response, api_key = requestor.request(method="post", url=url, params=wrapped_params)
        return convert_to_easypost_object(response=response, api_key=api_key)

    def buy(self, **params) -> "Batch":
        """Buy a batch."""
        requestor = Requestor(local_api_key=self._api_key)
        url = "%s/%s" % (self.instance_url(), "buy")
        response, api_key = requestor.request(method="post", url=url, params=params)
        self.refresh_from(values=response, api_key=api_key)
        return self

    def label(self, **params) -> "Batch":
        """Create a batch label."""
        requestor = Requestor(local_api_key=self._api_key)
        url = "%s/%s" % (self.instance_url(), "label")
        response, api_key = requestor.request(method="post", url=url, params=params)
        self.refresh_from(values=response, api_key=api_key)
        return self

    def remove_shipments(self, **params) -> "Batch":
        """Remove shipments from a batch."""
        requestor = Requestor(local_api_key=self._api_key)
        url = "%s/%s" % (self.instance_url(), "remove_shipments")
        response, api_key = requestor.request(method="post", url=url, params=params)
        self.refresh_from(values=response, api_key=api_key)
        return self

    def add_shipments(self, **params) -> "Batch":
        """Add shipments from a batch."""
        requestor = Requestor(local_api_key=self._api_key)
        url = "%s/%s" % (self.instance_url(), "add_shipments")
        response, api_key = requestor.request(method="post", url=url, params=params)
        self.refresh_from(values=response, api_key=api_key)
        return self

    def create_scan_form(self, **params) -> "Batch":
        """Create a scanform for a batch."""
        requestor = Requestor(local_api_key=self._api_key)
        url = "%s/%s" % (self.instance_url(), "scan_form")
        response, api_key = requestor.request(method="post", url=url, params=params)
        self.refresh_from(values=response, api_key=api_key)
        return self



from easypost.requestor import Requestor
from easypost.resource import (
    AllResource,
    CreateResource,
    DeleteResource,
)


class Webhook(AllResource, CreateResource, DeleteResource):
    def update(self, **params) -> "Webhook":
        """Update a webhook."""
        requestor = Requestor(local_api_key=self._api_key)
        url = self.instance_url()
        response, api_key = requestor.request(method="put", url=url, params=params)
        self.refresh_from(values=response, api_key=api_key)
        return self



import json
from typing import (
    Optional,
    Union,
)


class Error(Exception):
    def __init__(
        self,
        message: Optional[str] = None,
        http_status: Optional[int] = None,
        http_body: Optional[Union[str, bytes]] = None,
        original_exception: Optional[Exception] = None,
    ):
        super(Error, self).__init__(message)
        self.message = message
        self.http_status = http_status
        self.http_body = http_body
        self.original_exception = original_exception
        if http_body:
            try:
                self.json_body = json.loads(http_body)
            except Exception:
                self.json_body = None

        self.param = None
        try:
            self.param = self.json_body["error"].get("param", None)
        except Exception:
            pass



import datetime
import json
import platform
import ssl
import time
from json import JSONDecodeError
from typing import (
    Any,
    Dict,
    Optional,
    Tuple,
    Union,
)
from urllib.parse import urlencode

from easypost.constant import (
    SUPPORT_EMAIL,
    TIMEOUT,
    USER_AGENT,
)
from easypost.easypost_object import EasyPostObject
from easypost.error import Error


class Requestor:
    def __init__(self, local_api_key: Optional[str] = None):
        self._api_key = local_api_key

    @classmethod
    def _objects_to_ids(cls, param: Dict[str, Any]) -> Dict[str, Any]:
        """If providing an object as a parameter to another object,
        only pass along the ID so the API will use the object reference correctly."""
        if isinstance(param, EasyPostObject):
            return {"id": param.id}
        elif isinstance(param, dict):
            data = {}
            for (k, v) in param.items():
                if isinstance(v, list):
                    data[k] = [cls._objects_to_ids(item) for item in v]  # type: ignore
                else:
                    data[k] = cls._objects_to_ids(v)  # type: ignore
            return data
        else:
            return param

    def request(
        self,
        method: str,
        url: str,
        params: Optional[Dict[str, Any]] = None,
        api_key_required: bool = True,
    ) -> Tuple[dict, Optional[str]]:
        """Make a request to the EasyPost API."""
        if params is None:
            params = {}
        http_body, http_status, my_api_key = self.request_raw(
            method=method, url=url, params=params, api_key_required=api_key_required
        )
        response = self.interpret_response(http_body=http_body, http_status=http_status)
        return response, my_api_key

    def request_raw(
        self,
        method: str,
        url: str,
        params: Optional[Dict[str, Any]] = None,
        api_key_required: bool = True,
    ) -> Tuple[str, int, Optional[str]]:
        """Internal logic required to make a request to the EasyPost API."""
        # Importing here to avoid circular imports
        from easypost import (
            VERSION,
            api_base,
            api_key,
        )

        my_api_key = self._api_key or api_key

        if api_key_required and my_api_key is None:
            raise Error(
                "No API key provided. Set an API key via \"easypost.api_key = 'APIKEY'. "
                "Your API keys can be found in your EasyPost dashboard, or you can email us "
                f"at {SUPPORT_EMAIL} for assistance."
            )

        abs_url = "%s%s" % (api_base, url or "")
        params = self._objects_to_ids(param=params or {})

        ua = {
            "client_version": VERSION,
            "lang": "python",
            "publisher": "easypost",
            "request_lib": request_lib,
        }
        for attr, func in (
            ("lang_version", platform.python_version),
            ("platform", platform.platform),
            ("uname", lambda: " ".join(platform.uname())),
            ("implementation", platform.python_implementation),
        ):
            try:
                val = func()  # type: ignore
            except Exception as e:
                val = "!! %s" % e
            ua[attr] = val

        if hasattr(ssl, "OPENSSL_VERSION"):
            ua["openssl_version"] = ssl.OPENSSL_VERSION

        headers = {
            "X-Client-User-Agent": json.dumps(ua),
            "User-Agent": USER_AGENT,
            "Authorization": "Bearer %s" % my_api_key,
        }

        if request_lib == "urlfetch":
            http_body, http_status = self.urlfetch_request(
                method=method, abs_url=abs_url, headers=headers, params=params
            )
        elif request_lib == "requests":
            http_body, http_status = self.requests_request(
                method=method, abs_url=abs_url, headers=headers, params=params
            )
        else:
            raise Error(f"Bug discovered: invalid request_lib: {request_lib}. Please report to {SUPPORT_EMAIL}.")

        return http_body, http_status, my_api_key

    def interpret_response(self, http_body: str, http_status: int) -> Dict[str, Any]:
        """Interpret the response body we receive from the API."""
        if http_status == 204:
            # HTTP 204 does not have any response body and we can just return here
            return {}
        try:
            response = json.loads(http_body)
        except JSONDecodeError:
            raise Exception(
                f"Unable to parse response body from API: {http_body}\nHTTP response code was: {http_status}"
            )
        if not (200 <= http_status < 300):
            self.handle_api_error(http_status=http_status, http_body=http_body, response=response)
        return response

    def requests_request(
        self,
        method: str,
        abs_url: str,
        headers: Dict[str, Any],
        params: Dict[str, Any],
    ) -> Tuple[str, int]:
        """Make a request by using the `request` library."""
        method = method.lower()
        if method == "get" or method == "delete":
            url_params = params
            json = None
        elif method == "post" or method == "put":
            json = params
            url_params = None
        else:
            raise Error(f"Bug discovered: invalid request method: {method}. Please report to {SUPPORT_EMAIL}.")

        try:
            result = requests_session.request(
                method,
                abs_url,
                params=url_params,
                headers=headers,
                json=json,
                timeout=timeout,
                verify=True,
            )
            http_body = result.text
            http_status = result.status_code
        except Exception as e:
            raise Error(
                "Unexpected error communicating with EasyPost. If this "
                f"problem persists please let us know at {SUPPORT_EMAIL}.",
                original_exception=e,
            )
        return http_body, http_status

    def urlfetch_request(
        self,
        method: str,
        abs_url: str,
        headers: Dict[str, Any],
        params: Dict[str, Any],
    ) -> Tuple[str, int]:
        """Make a request by using the `urlfetch` library."""
        fetch_args = {"method": method, "headers": headers, "validate_certificate": False, "deadline": timeout}
        if method == "get" or method == "delete":
            fetch_args["url"] = self.add_params_to_url(url=abs_url, params=params)
        elif method == "post" or method == "put":
            fetch_args["url"] = abs_url
            fetch_args["payload"] = json.dumps(params, default=self._utf8)
        else:
            raise Error(f"Bug discovered: invalid request method: {method}. Please report to {SUPPORT_EMAIL}.")

        try:
            from google.appengine.api import urlfetch  # type: ignore

            result = urlfetch.fetch(**fetch_args)  # type: ignore
        except Exception as e:
            raise Error(
                "Unexpected error communicating with EasyPost. "
                f"If this problem persists, let us know at {SUPPORT_EMAIL}.",
                original_exception=e,
            )

        return result.content, result.status_code

    def handle_api_error(self, http_status: int, http_body: str, response: Dict[str, Any]) -> None:
        """Handles API errors returned from EasyPost."""
        try:
            error = response["error"]
        except (KeyError, TypeError):
            raise Error("Invalid response from API: (%d) %r " % (http_status, http_body), http_status, http_body)

        try:
            raise Error(message=error.get("message", ""), http_status=http_status, http_body=http_body)
        except AttributeError:
            raise Error(message=error, http_status=http_status, http_body=http_body)

    def _utf8(self, value: Union[str, bytes]) -> str:
        # Python3's str(bytestring) returns "b'bytestring'" so we use .decode instead
        if isinstance(value, bytes):
            return value.decode(encoding="utf-8")
        return value

    def encode_url_params(self, params: Dict[str, Any]) -> Union[str, None]:
        """Encode params for a URL."""
        converted_params = []
        for key, value in sorted(params.items()):
            if value is None:
                continue  # Don't add Nones to the query
            elif isinstance(value, datetime.datetime):
                value = int(time.mktime(value.timetuple()))  # to UTC timestamp
            converted_params.append((key, value))
        return urlencode(query=converted_params)

    def add_params_to_url(self, url: str, params: Dict[str, Any]) -> str:
        """Add params to the URL."""
        encoded_params = self.encode_url_params(params=params)
        if encoded_params:
            return "%s?%s" % (url, encoded_params)
        return url


# use urlfetch as request_lib on google app engine, otherwise use requests
request_lib = None
try:
    from google.appengine.api import urlfetch

    request_lib = "urlfetch"
    # use the GAE application-wide "deadline" (or its default)
    timeout = urlfetch.get_default_fetch_deadline() or TIMEOUT
except ImportError:
    timeout = TIMEOUT
    try:
        import requests

        request_lib = "requests"
        requests_session = requests.Session()
        requests_http_adapter = requests.adapters.HTTPAdapter(max_retries=3)
        requests_session.mount(prefix="https://api.easypost.com", adapter=requests_http_adapter)
    except Exception:
        raise ImportError(
            "EasyPost requires an up to date requests library. "
            'Update requests via "pip install -U requests" or '
            f"contact us at {SUPPORT_EMAIL}."
        )

    try:
        version = requests.__version__
        major, minor, patch = [int(i) for i in version.split(".")]
    except Exception:
        raise ImportError(
            "EasyPost requires an up to date requests library. "
            'Update requests via "pip install -U requests" or contact '
            f"us at {SUPPORT_EMAIL}."
        )
    else:
        if major < 1:
            raise ImportError(
                "EasyPost requires an up to date requests library. Update "
                'requests via "pip install -U requests" or contact us '
                f"at {SUPPORT_EMAIL}."
            )



from typing import (
    List,
    Optional,
)

from easypost.easypost_object import convert_to_easypost_object
from easypost.requestor import Requestor
from easypost.resource import (
    AllResource,
    CreateResource,
    DeleteResource,
    UpdateResource,
)


class CarrierAccount(AllResource, CreateResource, UpdateResource, DeleteResource):
    @classmethod
    def types(cls, api_key: Optional[str] = None) -> List[str]:
        """Get the types of carrier accounts available to the user."""
        requestor = Requestor(local_api_key=api_key)
        response, api_key = requestor.request(method="get", url="/carrier_types")
        return convert_to_easypost_object(response=response, api_key=api_key)



from easypost.requestor import Requestor
from easypost.resource import CreateResource


class Order(CreateResource):
    def get_rates(self) -> "Order":
        """Get rates for an order."""
        requestor = Requestor(local_api_key=self._api_key)
        url = "%s/%s" % (self.instance_url(), "rates")
        response, api_key = requestor.request(method="get", url=url)
        self.refresh_from(values=response, api_key=api_key)
        return self

    def buy(self, **params) -> "Order":
        """Buy an order."""
        requestor = Requestor(local_api_key=self._api_key)
        url = "%s/%s" % (self.instance_url(), "buy")
        response, api_key = requestor.request(method="post", url=url, params=params)
        self.refresh_from(values=response, api_key=api_key)
        return self



from easypost.requestor import Requestor
from easypost.resource import CreateResource


class Pickup(CreateResource):
    def buy(self, **params) -> "Pickup":
        """Buy a pickup."""
        requestor = Requestor(local_api_key=self._api_key)
        url = "%s/%s" % (self.instance_url(), "buy")
        response, api_key = requestor.request(method="post", url=url, params=params)
        self.refresh_from(values=response, api_key=api_key)
        return self

    def cancel(self, **params) -> "Pickup":
        """Cancel a pickup."""
        requestor = Requestor(local_api_key=self._api_key)
        url = "%s/%s" % (self.instance_url(), "cancel")
        response, api_key = requestor.request(method="post", url=url, params=params)
        self.refresh_from(values=response, api_key=api_key)
        return self



import json
from typing import (
    Any,
    Dict,
    List,
    Optional,
)

import easypost


EASYPOST_OBJECT_ID_PREFIX_TO_CLASS_NAME_MAP = {
    "adr": "Address",
    "batch": "Batch",
    "brd": "Brand",
    "ca": "CarrierAccount",
    "cfrep": "Report",
    "cstinfo": "CustomsInfo",
    "cstitem": "CustomsItem",
    "evt": "Event",
    "hook": "Webhook",
    "ins": "Insurance",
    "order": "Order",
    "pickup": "Pickup",
    "pickuprate": "PickupRate",
    "pl": "PostageLabel",
    "plrep": "Report",
    "prcl": "Parcel",
    "rate": "Rate",
    "refrep": "Report",
    "rfnd": "Refund",
    "sf": "ScanForm",
    "shp": "Shipment",
    "shpinvrep": "Report",
    "shprep": "Report",
    "trk": "Tracker",
    "trkrep": "Report",
    "user": "User",
}

OBJECT_CLASS_NAME_OVERRIDES = {
    "CashFlowReport": "Report",
    "PaymentLogReport": "Report",
    "RefundReport": "Report",
    "ShipmentInvoiceReport": "Report",
    "ShipmentReport": "Report",
    "TrackerReport": "Report",
}


def convert_to_easypost_object(
    response: Dict[str, Any],
    api_key: Optional[str] = None,
    parent: object = None,
    name: Optional[str] = None,
):
    """Convert a response to an EasyPost object."""
    if isinstance(response, list):
        return [convert_to_easypost_object(response=item, api_key=api_key, parent=parent) for item in response]
    elif isinstance(response, dict):
        response = response.copy()
        object_type_str = response.get("object", "EasyPostObject")
        class_name = OBJECT_CLASS_NAME_OVERRIDES.get(object_type_str, object_type_str)

        object_id = response.get("id", None)
        if object_id is not None:
            # If an object ID is present, use it to find the class type instead.
            object_id_prefix = object_id.split("_")[0]
            class_name = EASYPOST_OBJECT_ID_PREFIX_TO_CLASS_NAME_MAP.get(object_id_prefix, "EasyPostObject")

        cls = getattr(easypost, class_name, EasyPostObject)
        obj = cls.construct_from(values=response, api_key=api_key, parent=parent, name=name)
        return obj
    else:
        return response


class EasyPostObject(object):
    def __init__(
        self,
        easypost_id: Optional[str] = None,
        api_key: Optional[str] = None,
        parent: object = None,
        name: Optional[str] = None,
        **params,
    ):
        self.__dict__["_values"] = set()
        self.__dict__["_unsaved_values"] = set()
        self.__dict__["_transient_values"] = set()
        self.__dict__["_immutable_values"] = {"_api_key", "id"}
        self.__dict__["_retrieve_params"] = params

        self.__dict__["_parent"] = parent
        self.__dict__["_name"] = name

        self.__dict__["_api_key"] = api_key

        if easypost_id:
            self.id = easypost_id

    def __setattr__(self, k, v: Any) -> None:
        self.__dict__[k] = v

        if k not in self._immutable_values:
            self._values.add(k)
            self._unsaved_values.add(k)

            cur = self
            cur_parent = self._parent
            while cur_parent:
                if cur._name:
                    cur_parent._unsaved_values.add(cur._name)
                cur = cur_parent
                cur_parent = cur._parent

    def __getattr__(self, k) -> Any:
        try:
            return self.__dict__[k]
        except KeyError:
            pass
        raise AttributeError("%r object has no attribute %r" % (type(self).__name__, k))

    def __getitem__(self, k):
        return self.__dict__[k]

    def get(self, k, default: Any = None) -> Any:
        try:
            return self[k]
        except KeyError:
            return default

    def setdefault(self, k, default: Any = None) -> Any:
        try:
            return self[k]
        except KeyError:
            self[k] = default
        return default

    def __setitem__(self, k, v) -> None:
        setattr(self, k, v)

    @property
    def keys(self) -> List[str]:
        return self._values.keys()

    @property
    def values(self) -> List[Any]:
        return self._values.keys()

    @classmethod
    def construct_from(
        cls,
        values: Dict[str, Any],
        api_key: Optional[str] = None,
        parent: object = None,
        name: Optional[str] = None,
    ) -> object:
        """Construct an object."""
        instance = cls(easypost_id=values.get("id"), api_key=api_key, parent=parent, name=name)
        instance.refresh_from(values=values, api_key=api_key)
        return instance

    def refresh_from(self, values: Dict[str, Any], api_key: Optional[str] = None) -> None:
        """Update local object with changes from the API."""
        self._api_key = api_key

        for k, v in sorted(values.items()):
            if k == "id" and self.id != v:
                self.id = v
            if k in self._immutable_values:
                continue
            self.__dict__[k] = convert_to_easypost_object(response=v, api_key=api_key, parent=self, name=k)
            self._values.add(k)
            self._transient_values.discard(k)
            self._unsaved_values.discard(k)

    def flatten_unsaved(self) -> Dict[str, Any]:
        """Return a dict of `_unsaved_values` values from the current object."""
        values = {}
        for key in self._unsaved_values:
            value = self.get(key)
            if type(value) is EasyPostObject:
                values[key] = value.flatten_unsaved()
            else:
                values[key] = value
        return values

    def __repr__(self) -> str:
        type_string = ""

        if isinstance(self.get("object"), str):
            type_string = " %s" % self.get("object").encode("utf8")

        json_string = json.dumps(obj=self.to_dict(), sort_keys=True, indent=2, cls=EasyPostObjectEncoder)

        return "<%s%s at %s> JSON: %s" % (type(self).__name__, type_string, hex(id(self)), json_string)

    def __str__(self) -> str:
        return self.to_json(indent=2)

    def __eq__(self, other) -> bool:
        if not isinstance(other, EasyPostObject):
            return False
        return self.__str__() == other.__str__()

    def to_json(self, indent: int = None) -> str:
        """Convert current object to json string."""
        return json.dumps(obj=self.to_dict(), sort_keys=True, indent=indent, cls=EasyPostObjectEncoder)

    def to_dict(self) -> Dict[str, Any]:
        """Convert current object to a dict."""

        def _serialize(o):
            if isinstance(o, EasyPostObject):
                return o.to_dict()
            if isinstance(o, list):
                return [_serialize(r) for r in o]
            return o

        d = {"id": self.get("id")} if self.get("id") else {}
        for k in sorted(self._values):
            v = getattr(self, k)
            v = _serialize(v)
            d[k] = v
        return d


class EasyPostObjectEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        """Convert easypost object to a dict."""
        if isinstance(obj, EasyPostObject):
            return obj.to_dict()
        else:
            return json.JSONEncoder.default(self, o=obj)



from typing import List

from easypost import Rate
from easypost.error import Error
from easypost.requestor import Requestor
from easypost.resource import (
    AllResource,
    CreateResource,
)


class Shipment(AllResource, CreateResource):
    def regenerate_rates(self) -> "Shipment":
        """Regenerate rates for a shipment."""
        requestor = Requestor(local_api_key=self._api_key)
        url = "%s/%s" % (self.instance_url(), "rerate")
        response, api_key = requestor.request(method="post", url=url)
        self.refresh_from(values=response, api_key=api_key)
        return self

    def get_smartrates(self) -> List[object]:
        """Get smartrates for a shipment."""
        requestor = Requestor(local_api_key=self._api_key)
        url = "%s/%s" % (self.instance_url(), "smartrate")
        response, api_key = requestor.request(method="get", url=url)
        return response.get("result", [])

    def buy(self, **params) -> "Shipment":
        """Buy a shipment."""
        requestor = Requestor(local_api_key=self._api_key)
        url = "%s/%s" % (self.instance_url(), "buy")
        response, api_key = requestor.request(method="post", url=url, params=params)
        self.refresh_from(values=response, api_key=api_key)
        return self

    def refund(self, **params) -> "Shipment":
        """Refund a shipment."""
        requestor = Requestor(local_api_key=self._api_key)
        url = "%s/%s" % (self.instance_url(), "refund")
        response, api_key = requestor.request(method="post", url=url, params=params)
        self.refresh_from(values=response, api_key=api_key)
        return self

    def insure(self, **params) -> "Shipment":
        """Insure a shipment."""
        requestor = Requestor(local_api_key=self._api_key)
        url = "%s/%s" % (self.instance_url(), "insure")
        response, api_key = requestor.request(method="post", url=url, params=params)
        self.refresh_from(values=response, api_key=api_key)
        return self

    def label(self, **params) -> "Shipment":
        """Convert the label format of a shipment."""
        requestor = Requestor(local_api_key=self._api_key)
        url = "%s/%s" % (self.instance_url(), "label")
        response, api_key = requestor.request(method="get", url=url, params=params)
        self.refresh_from(values=response, api_key=api_key)
        return self

    def lowest_rate(self, carriers: List[str] = None, services: List[str] = None) -> Rate:
        """Get the lowest rate of a shipment."""
        carriers = carriers or []
        services = services or []
        lowest_rate = None

        carriers = [c.lower() for c in carriers]
        services = [service.lower() for service in services]

        for rate in self.rates:
            if len(carriers) > 0 and rate.carrier.lower() not in carriers:
                continue

            if len(services) > 0 and rate.service.lower() not in services:
                continue

            if lowest_rate is None or float(rate.rate) < float(lowest_rate.rate):
                lowest_rate = rate

        if lowest_rate is None:
            raise Error(message="No rates found.")

        return lowest_rate



from typing import (
    Any,
    Dict,
    Optional,
)

from easypost.easypost_object import convert_to_easypost_object
from easypost.requestor import Requestor
from easypost.resource import (
    AllResource,
    CreateResource,
)


class Address(AllResource, CreateResource):
    @classmethod
    def create(
        cls,
        api_key: Optional[str] = None,
        verify: Optional[Dict[str, Any]] = None,
        verify_strict: Dict[str, Any] = None,
        **params
    ) -> "Address":
        """Create an address."""
        requestor = Requestor(local_api_key=api_key)
        url = cls.class_url()

        wrapped_params = {cls.snakecase_name(): params}
        for key, value in (("verify", verify), ("verify_strict", verify_strict)):
            if not value:
                continue
            elif isinstance(value, (bool, str)):
                value = [value]
            wrapped_params[key] = value

        response, api_key = requestor.request(method="post", url=url, params=wrapped_params)
        return convert_to_easypost_object(response=response, api_key=api_key)

    @classmethod
    def create_and_verify(cls, api_key: Optional[str] = None, **params) -> "Address":
        """Create and verify an address."""
        requestor = Requestor(local_api_key=api_key)
        url = "%s/%s" % (cls.class_url(), "create_and_verify")

        wrapped_params = {cls.snakecase_name(): params}
        response, api_key = requestor.request(method="post", url=url, params=wrapped_params)

        response_address = response.get("address", None)

        if response_address is not None:
            verified_address = convert_to_easypost_object(response=response_address, api_key=api_key)
            return verified_address
        else:
            return convert_to_easypost_object(response=response, api_key=api_key)

    def verify(self, carrier: Optional[str] = None) -> "Address":
        """Verify an address."""
        requestor = Requestor(local_api_key=self._api_key)
        url = "%s/%s" % (self.instance_url(), "verify")
        params = {"carrier": carrier}
        response, api_key = requestor.request(method="get", url=url, params=params)

        response_address = response.get("address", None)
        response_message = response.get("message", None)

        if response_address is not None:
            verified_address = convert_to_easypost_object(response=response_address, api_key=api_key)
            if response_message is not None:
                verified_address.message = response_message
                verified_address._immutable_values.update("message")
            return verified_address
        else:
            return convert_to_easypost_object(response=response, api_key=api_key)



from typing import (
    List,
    Optional,
)

from easypost.easypost_object import convert_to_easypost_object
from easypost.requestor import Requestor
from easypost.resource import (
    CreateResource,
    DeleteResource,
    UpdateResource,
)


class User(CreateResource, UpdateResource, DeleteResource):
    @classmethod
    def create(cls, api_key: Optional[str] = None, **params) -> "User":
        """Create a user."""
        requestor = Requestor(local_api_key=api_key)
        url = cls.class_url()
        wrapped_params = {cls.snakecase_name(): params}
        response, api_key = requestor.request(method="post", url=url, params=wrapped_params, api_key_required=False)
        return convert_to_easypost_object(response=response, api_key=api_key)

    @classmethod
    def retrieve(cls, easypost_id: Optional[str] = None, api_key: Optional[str] = None, **params) -> "User":
        """Retrieve a user."""
        if not easypost_id:
            requestor = Requestor(local_api_key=api_key)
            response, api_key = requestor.request(method="get", url=cls.class_url())
            return convert_to_easypost_object(response=response, api_key=api_key)
        else:
            instance = cls(easypost_id=easypost_id, api_key=api_key, **params)
            instance.refresh()
            return instance

    @classmethod
    def retrieve_me(cls, api_key: Optional[str] = None, **params) -> "User":
        """Retrieve the authenticated user."""
        requestor = Requestor(local_api_key=api_key)
        response, api_key = requestor.request(method="get", url=cls.class_url())
        return convert_to_easypost_object(response=response, api_key=api_key)

    @classmethod
    def all_api_keys(cls, api_key: Optional[str] = None) -> "User":
        """Get all API keys including child user keys."""
        requestor = Requestor(local_api_key=api_key)
        url = "/api_keys"
        response, api_key = requestor.request(method="get", url=url)
        return convert_to_easypost_object(response=response, api_key=api_key)

    def api_keys(self) -> List[str]:
        """Get the authenticated user's API keys."""
        api_keys = self.all_api_keys()

        if api_keys.id == self.id:
            return api_keys.keys
        else:
            for child in api_keys.children:
                if child.id == self.id:
                    return child.keys

        return []

    def update_brand(self, api_key: Optional[str] = None, **params) -> "User":
        """Update the User's Brand."""
        requestor = Requestor(local_api_key=api_key)
        response, api_key = requestor.request(method="put", url=self.instance_url() + "/brand", params=params)
        return convert_to_easypost_object(response=response, api_key=api_key)



from setuptools import setup


REQUIREMENTS = [
    "requests >= 2.4.3",
    "typing-extensions",
]

DEV_REQUIREMENTS = [
    "black",
    "flake8",
    "isort",
    "pytest-cov==3.*",
    "pytest-vcr==1.*",
    "pytest==7.*",
    "types-requests",
    "types-urllib3",
    "vcrpy==4.*",
]

# packages incompatible with PyPy go here
CPYTHON_DEV_REQUIREMENTS = [
    "mypy",
]

RELEASE_REQUIREMENTS = [
    "twine==3.8.*",  # Twine 4.0+ drops support for Python 3.6
    "wheel",
]

with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="easypost",
    version="7.0.0",
    description="EasyPost Shipping API Client Library for Python",
    author="EasyPost",
    author_email="support@easypost.com",
    url="https://easypost.com/",
    packages=["easypost"],
    install_requires=REQUIREMENTS,
    extras_require={
        "dev": DEV_REQUIREMENTS + CPYTHON_DEV_REQUIREMENTS,
        "pypy_dev": DEV_REQUIREMENTS,  # no cpython requirements
        "release": RELEASE_REQUIREMENTS,
    },
    package_data={
        "easypost": ["py.typed"],
    },
    test_suite="test",
    long_description=long_description,
    long_description_content_type="text/markdown",
    project_urls={
        "Docs": "https://www.easypost.com/docs/api",
        "Tracker": "https://github.com/EasyPost/easypost-python/issues",
        "Source": "https://github.com/EasyPost/easypost-python",
    },
    python_requires=">=3.6, <4",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries",
    ],
)



import re
from typing import (
    Any,
    List,
    Optional,
)

from easypost.easypost_object import (
    EasyPostObject,
    convert_to_easypost_object,
)
from easypost.error import Error
from easypost.requestor import Requestor


class Resource(EasyPostObject):
    def _ident(self) -> List[str]:
        return [self.get("id")]

    @classmethod
    def retrieve(cls, easypost_id: str, api_key: Optional[str] = None, **params) -> object:
        """Retrieve an object from the EasyPost API."""
        instance = cls(easypost_id=easypost_id, api_key=api_key, **params)
        instance.refresh()
        return instance

    def refresh(self) -> object:
        """Refresh the local object from the API response."""
        requestor = Requestor(local_api_key=self._api_key)
        url = self.instance_url()
        response, api_key = requestor.request(method="get", url=url, params=self._retrieve_params)
        self.refresh_from(values=response, api_key=api_key)
        return self

    @classmethod
    def snakecase_name(cls) -> str:
        """Return the class name as snake_case."""
        snake_name = (cls.__name__[0:1] + re.sub(r"([A-Z])", r"_\1", cls.__name__[1:])).lower()
        return snake_name

    @classmethod
    def class_url(cls) -> str:
        """Generate a URL based on class name."""
        cls_name = cls.snakecase_name()
        if cls_name[-1:] in ("s", "h"):
            return "/%ses" % cls_name
        else:
            return "/%ss" % cls_name

    def instance_url(self) -> str:
        """Generate an instance URL based on the ID of the object."""
        easypost_id = self.get("id")
        if not easypost_id:
            raise Error("%s instance has invalid ID: %r" % (type(self).__name__, easypost_id))
        return "%s/%s" % (self.class_url(), easypost_id)


class AllResource(Resource):
    @classmethod
    def all(cls, api_key: Optional[str] = None, **params) -> List[Any]:
        """Retrieve a list of records from the API."""
        requestor = Requestor(local_api_key=api_key)
        url = cls.class_url()
        response, api_key = requestor.request(method="get", url=url, params=params)
        return convert_to_easypost_object(response=response, api_key=api_key)


class CreateResource(Resource):
    @classmethod
    def create(cls, api_key: Optional[str] = None, **params) -> Any:
        """Create an EasyPost object."""
        requestor = Requestor(local_api_key=api_key)
        url = cls.class_url()
        wrapped_params = {cls.snakecase_name(): params}
        response, api_key = requestor.request(method="post", url=url, params=wrapped_params)
        return convert_to_easypost_object(response=response, api_key=api_key)


class UpdateResource(Resource):
    def save(self) -> Any:
        """Update an EasyPost object."""
        if self._unsaved_values:
            requestor = Requestor(local_api_key=self._api_key)
            params = {}
            for k in self._unsaved_values:
                params[k] = getattr(self, k)
                if type(params[k]) is EasyPostObject:
                    params[k] = params[k].flatten_unsaved()
            params = {self.snakecase_name(): params}
            url = self.instance_url()
            response, api_key = requestor.request(method="put", url=url, params=params)
            self.refresh_from(values=response, api_key=api_key)

        return self


class DeleteResource(Resource):
    def delete(self, **params) -> Any:
        """Delete an EasyPost object."""
        requestor = Requestor(local_api_key=self._api_key)
        url = self.instance_url()
        response, api_key = requestor.request(method="delete", url=url, params=params)
        self.refresh_from(values=response, api_key=api_key)
        return self



from easypost.address import Address
from easypost.batch import Batch
from easypost.brand import Brand
from easypost.carrier_account import CarrierAccount
from easypost.customs_info import CustomsInfo
from easypost.customs_item import CustomsItem
from easypost.error import Error
from easypost.event import Event
from easypost.insurance import Insurance
from easypost.order import Order
from easypost.parcel import Parcel
from easypost.pickup import Pickup
from easypost.pickup_rate import PickupRate
from easypost.postage_label import PostageLabel
from easypost.rate import Rate
from easypost.refund import Refund
from easypost.report import Report
from easypost.scanform import ScanForm
from easypost.shipment import Shipment
from easypost.tax_identifier import TaxIdentifier
from easypost.tracker import Tracker
from easypost.user import User
from easypost.webhook import Webhook

from .version import (
    VERSION,
    VERSION_INFO,
)


__author__ = "EasyPost <oss@easypost.com>"
__version__ = VERSION
version_info = VERSION_INFO

# config
api_key = None
api_base = "https://api.easypost.com/v2"



from typing import Optional

from easypost.easypost_object import convert_to_easypost_object
from easypost.requestor import Requestor
from easypost.resource import (
    AllResource,
    CreateResource,
)


class Report(AllResource, CreateResource):
    @classmethod
    def create(cls, api_key: Optional[str] = None, **params):
        """Create a report."""
        requestor = Requestor(local_api_key=api_key)
        url = f"{cls.class_url()}/{params.get('type')}"
        response, api_key = requestor.request(method="post", url=url, params=params)
        return convert_to_easypost_object(response=response, api_key=api_key)

    @classmethod
    def all(cls, api_key: Optional[str] = None, **params):
        """Retrieve all reports."""
        requestor = Requestor(local_api_key=api_key)
        url = f"{cls.class_url()}/{params.get('type')}"
        response, api_key = requestor.request(method="get", url=url, params=params)
        return convert_to_easypost_object(response=response, api_key=api_key)
